# Copyright 2016-2018 The NATS Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import asyncio
import logging
import random
import stan.pb.protocol_pb2 as protocol
from stan.aio.errors import *
from time import time as now

__version__ = '0.3.0'

# Subject namespaces for clients to ack and connect
DEFAULT_ACKS_SUBJECT = "_STAN.acks.%s"
DEFAULT_DISCOVER_SUBJECT = "_STAN.discover.%s"
DEFAULT_INBOX_SUBJECT = "_INBOX.%s"

# Ack timeout in seconds
DEFAULT_ACK_WAIT = 30

# Max number of inflight acks from received messages
DEFAULT_MAX_INFLIGHT = 1024

# Connect timeout in seconds
DEFAULT_CONNECT_TIMEOUT = 2

# Max number of inflight pub acks
DEFAULT_MAX_PUB_ACKS_INFLIGHT = 16384

# Max number of pending messages awaiting
# to be processed on a single subscriptions.
DEFAULT_PENDING_LIMIT = 8192

logger = logging.getLogger(__name__)

class Client:
    """
    Asyncio Client for NATS Streaming.
    """

    def __init__(self):
        # NATS transport.
        self._nc = None
        self._loop = None
        self._nats_conn_is_borrowed = False

        # Options
        self._connect_timeout = None
        self._max_pub_acks_inflight = None

        # Inbox subscription for periodical heartbeat messages.
        self._hb_inbox = None
        self._hb_inbox_sid = None

        # Subscription for processing received acks from the server.
        self._ack_subject = None
        self._ack_subject_sid = None

        # Publish prefix set by stan to which we append our subject on publish.
        self._pub_prefix = None
        self._sub_req_subject = None
        self._unsub_req_subject = None
        self._close_req_subject = None
        self._sub_close_req_subject = None

        # Map of guid to futures which are signaled when the ack is processed.
        self._pub_ack_map = {}

        # Map of subscriptions related to the NATS Streaming session.
        self._sub_map = {}

    def __repr__(self):
        return "<nats streaming client v{}>".format(__version__)

    async def connect(self, cluster_id, client_id,
                      nats=None,
                      connect_timeout=DEFAULT_CONNECT_TIMEOUT,
                      max_pub_acks_inflight=DEFAULT_MAX_PUB_ACKS_INFLIGHT,
                      loop=None,
                      ):
        """
        Starts a session with a NATS Streaming cluster.

        :param cluster: Name of the cluster to which we will connect.
        :param nats: NATS connection to be borrowed for NATS Streaming.
        """
        self._cluster_id = cluster_id
        self._client_id = client_id
        self._loop = loop
        self._connect_timeout = connect_timeout

        if nats is not None:
            self._nc = nats
            # NATS Streaming client should use same event loop
            # as the borrowed NATS connection.
            self._loop = self._nc._loop

        # Subjects
        self._discover_subject = DEFAULT_DISCOVER_SUBJECT % self._cluster_id
        self._hb_inbox = DEFAULT_INBOX_SUBJECT % new_guid()
        self._ack_subject = DEFAULT_ACKS_SUBJECT % new_guid()

        # Pending pub acks inflight
        self._pending_pub_acks_queue = asyncio.Queue(
            maxsize=max_pub_acks_inflight, loop=self._loop)

        # Heartbeats subscription
        self._hb_inbox_sid = await self._nc.subscribe(
            self._hb_inbox,
            cb=self._process_heartbeats,
            )

        # Acks processing subscription
        self._ack_subject_sid = await self._nc.subscribe(
            self._ack_subject,
            cb=self._process_ack,
            )
        await self._nc.flush()

        # Start NATS Streaming session by sending ConnectRequest
        creq = protocol.ConnectRequest()
        creq.clientID = self._client_id
        creq.heartbeatInbox = self._hb_inbox
        payload = creq.SerializeToString()

        msg = None
        try:
            msg = await self._nc.request(
                self._discover_subject,
                payload,
                timeout=self._connect_timeout,
                )
        except:
            await self._close()
            raise ErrConnectReqTimeout("stan: failed connecting to '{}'".format(cluster_id))

        # We should get the NATS Streaming subject from the
        # response from the ConnectRequest.
        resp = protocol.ConnectResponse()
        resp.ParseFromString(msg.data)
        if resp.error != "":
            try:
                await self._close()
            except:
                pass
            raise StanError(resp.error)

        self._pub_prefix = resp.pubPrefix
        self._sub_req_subject = resp.subRequests
        self._unsub_req_subject = resp.unsubRequests
        self._close_req_subject = resp.closeRequests
        self._sub_close_req_subject = resp.subCloseRequests

    async def _process_heartbeats(self, msg):
        """
        Receives heartbeats sent to the client by the server.
        """
        await self._nc.publish(msg.reply, b'')

    async def _process_ack(self, msg):
        """
        Receives acks from the publishes via the _STAN.acks subscription.
        """
        pub_ack = protocol.PubAck()
        pub_ack.ParseFromString(msg.data)

        # Unblock pending acks queue if required.
        if not self._pending_pub_acks_queue.empty():
            await self._pending_pub_acks_queue.get()

        try:
            cb = self._pub_ack_map[pub_ack.guid]
            await cb(pub_ack)
            del self._pub_ack_map[pub_ack.guid]
        except KeyError:
            # Just skip the pub ack
            return
        except:
            # TODO: Check for protocol error
            return

    async def _process_msg(self, sub):
        """
        Receives the msgs from the STAN subscriptions and replies.
        By default it will reply back with an ack unless manual acking
        was specified in one of the subscription options.
        """
        while True:
            try:
                raw_msg = await sub._msgs_queue.get()
                msg = Msg()
                msg_proto = protocol.MsgProto()
                msg_proto.ParseFromString(raw_msg.data)
                msg.proto = msg_proto
                msg.sub = sub

                # Yield the message to the subscription callback.
                await sub.cb(msg)

                if not sub.manual_acks:
                    # Process auto-ack if not done manually in the callback,
                    # by publishing into the ack inbox from the subscription.
                    msg_ack = protocol.Ack()
                    msg_ack.subject = msg.proto.subject
                    msg_ack.sequence = msg.proto.sequence
                    await self._nc.publish(sub.ack_inbox, msg_ack.SerializeToString())
            except asyncio.CancelledError:
                break
            except Exception as ex:
                if sub.error_cb:
                    try:
                        await sub.error_cb(ex)
                    except:
                        logger.exception(
                            "Exception in error callback for subscription to '%s'",
                            sub.subject
                        )
                continue

    async def ack(self, msg):
        """
        Used to manually acks a message.

        :param msg: Message which is pending to be acked by client.
        """
        ack_proto = protocol.Ack()
        ack_proto.subject = msg.proto.subject
        ack_proto.sequence = msg.proto.sequence
        await self._nc.publish(msg.sub.ack_inbox, ack_proto.SerializeToString())

    async def publish(self, subject, payload,
                      ack_handler=None,
                      ack_wait=DEFAULT_ACK_WAIT,
                      ):
        """
        Publishes a payload onto a subject.  By default, it will block
        until the message which has been published has been acked back.
        An optional async handler can be publi

        :param subject: Subject of the message.
        :param payload: Payload of the message which wil be published.
        :param ack_handler: Optional handler for async publishing.
        :param ack_wait: How long in seconds to wait for an ack to be received.
        """
        stan_subject = ''.join([self._pub_prefix, '.', subject])
        guid = new_guid()
        pe = protocol.PubMsg()
        pe.clientID = self._client_id
        pe.guid = guid
        pe.subject = subject
        pe.data = payload

        # Control max inflight pubs for the client with a buffered queue.
        await self._pending_pub_acks_queue.put(None)

        # Process asynchronously if a handler is given.
        if ack_handler is not None:
            self._pub_ack_map[guid] = ack_handler

            try:
                await self._nc.publish_request(
                    stan_subject,
                    self._ack_subject,
                    pe.SerializeToString(),
                    )
                return
            except Exception as e:
                del self._pub_ack_map[guid]
                raise e
        else:
            # Synchronous wait for ack handling.
            future = asyncio.Future(loop=self._loop)
            async def cb(pub_ack):
                nonlocal future
                future.set_result(pub_ack)
            self._pub_ack_map[guid] = cb

            try:
                await self._nc.publish_request(
                    stan_subject,
                    self._ack_subject,
                    pe.SerializeToString(),
                    )
                await asyncio.wait_for(future, ack_wait, loop=self._loop)
                return future.result()
            except Exception as e:
                # Remove pending future before raising error.
                future.cancel()
                del self._pub_ack_map[guid]
                raise e

    async def subscribe(self, subject,
                        cb=None,
                        error_cb=None,
                        start_at=None,
                        deliver_all_available=False,
                        sequence=None,
                        time=None,
                        manual_acks=False,
                        queue=None,
                        ack_wait=DEFAULT_ACK_WAIT,
                        max_inflight=DEFAULT_MAX_INFLIGHT,
                        durable_name=None,
                        pending_limits=DEFAULT_PENDING_LIMIT,
                        ):
        """
        :param subject: Subject for the NATS Streaming subscription.

        :param cb: Callback which will be dispatched the

        :param error_cb: Async callback called on error, with the exception as
        the sole argument.

        :param start_at: One of the following options:
           - 'new_only' (default)
           - 'first'
           - 'sequence'
           - 'last_received'
           - 'time'

        :param deliver_all_available: Signals to receive all messages.

        :param sequence: Start sequence position from which we will be
        receiving the messages.

        :param time: Unix timestamp after which the messages will be delivered.

        :param manual_acks: Disables auto ack functionality in the subscription
        callback so that it is implemented by the user instead.

        :param ack_wait: How long to wait for an ack before being redelivered
        previous messages.

        :param durable_name: Name of the durable subscription.

        :param: pending_limits: Max number of messages to await in subscription
        before it becomes a slow consumer.
        """
        sub = Subscription(
            subject=subject,
            queue=queue,
            cb=cb,
            error_cb=error_cb,
            manual_acks=manual_acks,
            stan=self,
            )
        self._sub_map[sub.inbox] = sub

        # Have the message processing queue ready before making the subscription.
        sub._msgs_queue = asyncio.Queue(maxsize=pending_limits, loop=self._loop)

        # Helper coroutine which will just put messages in to the queue,
        # whenever the NATS client receives a message.
        async def cb(raw_msg):
            nonlocal sub
            await sub._msgs_queue.put(raw_msg)

        # Should create the NATS Subscription before making the request.
        sid = await self._nc.subscribe(sub.inbox, cb=cb)
        sub.sid = sid

        req = protocol.SubscriptionRequest()
        req.clientID = self._client_id
        req.maxInFlight = max_inflight
        req.ackWaitInSecs = ack_wait

        if queue is not None:
            req.qGroup = queue

        if durable_name is not None:
            req.durableName = durable_name

        # Normalize start position options.
        if deliver_all_available:
            req.startPosition = protocol.First
        elif start_at is None or start_at == 'new_only':
            req.startPosition = protocol.NewOnly
        elif start_at == 'last_received':
            req.startPosition = protocol.LastReceived
        elif start_at == 'time':
            req.startPosition = protocol.TimeDeltaStart
            req.startTimeDelta = int(now() - time) * 1000000000
        elif start_at == 'sequence':
            req.startPosition = protocol.SequenceStart
            req.startSequence = sequence
        elif start_at == 'first':
            req.startPosition = protocol.First

        # Set STAN subject and NATS inbox where we will be awaiting
        # for the messages to be delivered.
        req.subject = subject
        req.inbox = sub.inbox

        msg = await self._nc.request(
            self._sub_req_subject,
            req.SerializeToString(),
            self._connect_timeout,
            )
        resp = protocol.SubscriptionResponse()
        resp.ParseFromString(msg.data)

        # If there is an error here, then rollback the
        # subscription which we have sent already.
        if resp.error != "":
            try:
                await self._nc.unsubscribe(sid)
            except:
                pass
            raise StanError(resp.error)
        sub.ack_inbox = resp.ackInbox
        sub._msgs_task = self._loop.create_task(self._process_msg(sub))

        return sub

    async def _close(self):
        """
        Removes any present internal state from the client.
        """

        # Remove the core NATS Streaming subscriptions.
        try:
            if self._hb_inbox_sid is not None:
                await self._nc.unsubscribe(self._hb_inbox_sid)
                self._hb_inbox = None
                self._hb_inbox_sid = None
            if self._ack_subject_sid is not None:
                await self._nc.unsubscribe(self._ack_subject_sid)
                self._ack_subject = None
                self._ack_subject_sid = None
        except:
            # FIXME: async error in case these fail?
            pass

        # Remove all the related subscriptions
        for _, sub in self._sub_map.items():
            if sub._msgs_task is not None:
                sub._msgs_task.cancel()
            try:
                await self._nc.unsubscribe(sub.sid)
            except:
                continue
        self._sub_map = {}

    async def close(self):
        """
        Close terminates a session with NATS Streaming.
        """

        # Remove the core NATS Streaming subscriptions.
        await self._close()

        req = protocol.CloseRequest()
        req.clientID = self._client_id
        msg = await self._nc.request(
            self._close_req_subject,
            req.SerializeToString(),
            self._connect_timeout,
            )
        resp = protocol.CloseResponse()
        resp.ParseFromString(msg.data)

        if resp.error != "":
            raise StanError(resp.error)

class Subscription(object):

    def __init__(self,
                 subject='',
                 queue='',
                 cb=None,
                 error_cb=None,
                 sid=None,
                 durable_name=None,
                 ack_inbox=None,
                 manual_acks=False,
                 stan=None,
                 msgs_queue=None,
                 msgs_task=None,
                 ):
        self.subject = subject
        self.queue = queue
        self.cb = cb
        self.error_cb = error_cb
        self.inbox = DEFAULT_INBOX_SUBJECT % new_guid()
        self.sid = sid
        self.ack_inbox = ack_inbox
        self.durable_name = durable_name
        self.manual_acks = manual_acks
        self._sc = stan
        self._nc = stan._nc
        self._msgs_queue = msgs_queue
        self._msgs_task = msgs_task

    @property
    def pending_queue_size(self):
        return self._msgs_queue.qsize()

    async def unsubscribe(self):
        """
        Remove subscription on a topic in this client.
        """
        await self._nc.unsubscribe(self.sid)

        try:
            # Stop the processing task for the subscription.
            sub = self._sc._sub_map[self.inbox]
            sub._msgs_task.cancel()
            del self._sc._sub_map[self.inbox]
        except KeyError:
            pass

        req = protocol.UnsubscribeRequest()
        req.clientID = self._sc._client_id
        req.subject = self.subject
        req.inbox = self.ack_inbox

        if self.durable_name is not None:
            req.durableName = self.durable_name

        msg = await self._nc.request(
            self._sc._unsub_req_subject,
            req.SerializeToString(),
            self._sc._connect_timeout,
            )
        resp = protocol.SubscriptionResponse()
        resp.ParseFromString(msg.data)

        if resp.error != "":
            raise StanError(resp.error)

    async def close(self):
        """
        Closes a NATS streaming subscription.
        """
        await self._nc.unsubscribe(self.sid)

        try:
            # Stop the processing task for the subscription.
            sub = self._sc._sub_map[self.inbox]
            sub._msgs_task.cancel()
            del self._sc._sub_map[self.inbox]
        except KeyError:
            pass

        req = protocol.UnsubscribeRequest()
        req.clientID = self._sc._client_id
        req.subject = self.subject
        req.inbox = self.ack_inbox

        if self.durable_name is not None:
            req.durableName = self.durable_name

        msg = await self._nc.request(
            self._sc._sub_close_req_subject,
            req.SerializeToString(),
            self._sc._connect_timeout,
            )
        resp = protocol.SubscriptionResponse()
        resp.ParseFromString(msg.data)

        if resp.error != "":
            raise StanError(resp.error)

class Msg(object):

    def __init__(self):
        self.proto = None
        self.sub = None

    @property
    def data(self):
        return self.proto.data

    @property
    def sequence(self):
        return self.proto.sequence

    @property
    def seq(self):
        return self.proto.sequence

    @property
    def timestamp(self):
        return self.proto.timestamp

    def __repr__(self):
        return "<nats streaming msg sequence={}, time={}>".format(self.proto.sequence, self.proto.timestamp)

def new_guid():
    return "%x" % random.SystemRandom().getrandbits(0x58)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protocol.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protocol.proto',
  package='pb',
  syntax='proto3',
  serialized_pb=_b('\n\x0eprotocol.proto\x12\x02pb\"f\n\x06PubMsg\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\x12\x0c\n\x04guid\x18\x02 \x01(\t\x12\x0f\n\x07subject\x18\x03 \x01(\t\x12\r\n\x05reply\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\x0e\n\x06sha256\x18\n \x01(\x0c\"%\n\x06PubAck\x12\x0c\n\x04guid\x18\x01 \x01(\t\x12\r\n\x05\x65rror\x18\x02 \x01(\t\"\x81\x01\n\x08MsgProto\x12\x10\n\x08sequence\x18\x01 \x01(\x04\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\r\n\x05reply\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x11\n\ttimestamp\x18\x05 \x01(\x03\x12\x13\n\x0bredelivered\x18\x06 \x01(\x08\x12\r\n\x05\x43RC32\x18\n \x01(\r\"(\n\x03\x41\x63k\x12\x0f\n\x07subject\x18\x01 \x01(\t\x12\x10\n\x08sequence\x18\x02 \x01(\x04\":\n\x0e\x43onnectRequest\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\x12\x16\n\x0eheartbeatInbox\x18\x02 \x01(\t\"\xa3\x01\n\x0f\x43onnectResponse\x12\x11\n\tpubPrefix\x18\x01 \x01(\t\x12\x13\n\x0bsubRequests\x18\x02 \x01(\t\x12\x15\n\runsubRequests\x18\x03 \x01(\t\x12\x15\n\rcloseRequests\x18\x04 \x01(\t\x12\r\n\x05\x65rror\x18\x05 \x01(\t\x12\x18\n\x10subCloseRequests\x18\x06 \x01(\t\x12\x11\n\tpublicKey\x18\x64 \x01(\t\"\xf1\x01\n\x13SubscriptionRequest\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\x0e\n\x06qGroup\x18\x03 \x01(\t\x12\r\n\x05inbox\x18\x04 \x01(\t\x12\x13\n\x0bmaxInFlight\x18\x05 \x01(\x05\x12\x15\n\rackWaitInSecs\x18\x06 \x01(\x05\x12\x13\n\x0b\x64urableName\x18\x07 \x01(\t\x12(\n\rstartPosition\x18\n \x01(\x0e\x32\x11.pb.StartPosition\x12\x15\n\rstartSequence\x18\x0b \x01(\x04\x12\x16\n\x0estartTimeDelta\x18\x0c \x01(\x03\"7\n\x14SubscriptionResponse\x12\x10\n\x08\x61\x63kInbox\x18\x02 \x01(\t\x12\r\n\x05\x65rror\x18\x03 \x01(\t\"[\n\x12UnsubscribeRequest\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\x12\x0f\n\x07subject\x18\x02 \x01(\t\x12\r\n\x05inbox\x18\x03 \x01(\t\x12\x13\n\x0b\x64urableName\x18\x04 \x01(\t\" \n\x0c\x43loseRequest\x12\x10\n\x08\x63lientID\x18\x01 \x01(\t\"\x1e\n\rCloseResponse\x12\r\n\x05\x65rror\x18\x01 \x01(\t*`\n\rStartPosition\x12\x0b\n\x07NewOnly\x10\x00\x12\x10\n\x0cLastReceived\x10\x01\x12\x12\n\x0eTimeDeltaStart\x10\x02\x12\x11\n\rSequenceStart\x10\x03\x12\t\n\x05\x46irst\x10\x04\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_STARTPOSITION = _descriptor.EnumDescriptor(
  name='StartPosition',
  full_name='pb.StartPosition',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='NewOnly', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='LastReceived', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TimeDeltaStart', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SequenceStart', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='First', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1025,
  serialized_end=1121,
)
_sym_db.RegisterEnumDescriptor(_STARTPOSITION)

StartPosition = enum_type_wrapper.EnumTypeWrapper(_STARTPOSITION)
NewOnly = 0
LastReceived = 1
TimeDeltaStart = 2
SequenceStart = 3
First = 4



_PUBMSG = _descriptor.Descriptor(
  name='PubMsg',
  full_name='pb.PubMsg',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='pb.PubMsg.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='guid', full_name='pb.PubMsg.guid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subject', full_name='pb.PubMsg.subject', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reply', full_name='pb.PubMsg.reply', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data', full_name='pb.PubMsg.data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sha256', full_name='pb.PubMsg.sha256', index=5,
      number=10, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=124,
)


_PUBACK = _descriptor.Descriptor(
  name='PubAck',
  full_name='pb.PubAck',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='guid', full_name='pb.PubAck.guid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error', full_name='pb.PubAck.error', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=126,
  serialized_end=163,
)


_MSGPROTO = _descriptor.Descriptor(
  name='MsgProto',
  full_name='pb.MsgProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='sequence', full_name='pb.MsgProto.sequence', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subject', full_name='pb.MsgProto.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='reply', full_name='pb.MsgProto.reply', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='data', full_name='pb.MsgProto.data', index=3,
      number=4, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='pb.MsgProto.timestamp', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='redelivered', full_name='pb.MsgProto.redelivered', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='CRC32', full_name='pb.MsgProto.CRC32', index=6,
      number=10, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=295,
)


_ACK = _descriptor.Descriptor(
  name='Ack',
  full_name='pb.Ack',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='subject', full_name='pb.Ack.subject', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='sequence', full_name='pb.Ack.sequence', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=297,
  serialized_end=337,
)


_CONNECTREQUEST = _descriptor.Descriptor(
  name='ConnectRequest',
  full_name='pb.ConnectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='pb.ConnectRequest.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='heartbeatInbox', full_name='pb.ConnectRequest.heartbeatInbox', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=339,
  serialized_end=397,
)


_CONNECTRESPONSE = _descriptor.Descriptor(
  name='ConnectResponse',
  full_name='pb.ConnectResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='pubPrefix', full_name='pb.ConnectResponse.pubPrefix', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subRequests', full_name='pb.ConnectResponse.subRequests', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='unsubRequests', full_name='pb.ConnectResponse.unsubRequests', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='closeRequests', full_name='pb.ConnectResponse.closeRequests', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error', full_name='pb.ConnectResponse.error', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subCloseRequests', full_name='pb.ConnectResponse.subCloseRequests', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='publicKey', full_name='pb.ConnectResponse.publicKey', index=6,
      number=100, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=400,
  serialized_end=563,
)


_SUBSCRIPTIONREQUEST = _descriptor.Descriptor(
  name='SubscriptionRequest',
  full_name='pb.SubscriptionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='pb.SubscriptionRequest.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subject', full_name='pb.SubscriptionRequest.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='qGroup', full_name='pb.SubscriptionRequest.qGroup', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inbox', full_name='pb.SubscriptionRequest.inbox', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='maxInFlight', full_name='pb.SubscriptionRequest.maxInFlight', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ackWaitInSecs', full_name='pb.SubscriptionRequest.ackWaitInSecs', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='durableName', full_name='pb.SubscriptionRequest.durableName', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='startPosition', full_name='pb.SubscriptionRequest.startPosition', index=7,
      number=10, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='startSequence', full_name='pb.SubscriptionRequest.startSequence', index=8,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='startTimeDelta', full_name='pb.SubscriptionRequest.startTimeDelta', index=9,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=566,
  serialized_end=807,
)


_SUBSCRIPTIONRESPONSE = _descriptor.Descriptor(
  name='SubscriptionResponse',
  full_name='pb.SubscriptionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='ackInbox', full_name='pb.SubscriptionResponse.ackInbox', index=0,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error', full_name='pb.SubscriptionResponse.error', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=809,
  serialized_end=864,
)


_UNSUBSCRIBEREQUEST = _descriptor.Descriptor(
  name='UnsubscribeRequest',
  full_name='pb.UnsubscribeRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='pb.UnsubscribeRequest.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='subject', full_name='pb.UnsubscribeRequest.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='inbox', full_name='pb.UnsubscribeRequest.inbox', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='durableName', full_name='pb.UnsubscribeRequest.durableName', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=866,
  serialized_end=957,
)


_CLOSEREQUEST = _descriptor.Descriptor(
  name='CloseRequest',
  full_name='pb.CloseRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='clientID', full_name='pb.CloseRequest.clientID', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=959,
  serialized_end=991,
)


_CLOSERESPONSE = _descriptor.Descriptor(
  name='CloseResponse',
  full_name='pb.CloseResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='pb.CloseResponse.error', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=993,
  serialized_end=1023,
)

_SUBSCRIPTIONREQUEST.fields_by_name['startPosition'].enum_type = _STARTPOSITION
DESCRIPTOR.message_types_by_name['PubMsg'] = _PUBMSG
DESCRIPTOR.message_types_by_name['PubAck'] = _PUBACK
DESCRIPTOR.message_types_by_name['MsgProto'] = _MSGPROTO
DESCRIPTOR.message_types_by_name['Ack'] = _ACK
DESCRIPTOR.message_types_by_name['ConnectRequest'] = _CONNECTREQUEST
DESCRIPTOR.message_types_by_name['ConnectResponse'] = _CONNECTRESPONSE
DESCRIPTOR.message_types_by_name['SubscriptionRequest'] = _SUBSCRIPTIONREQUEST
DESCRIPTOR.message_types_by_name['SubscriptionResponse'] = _SUBSCRIPTIONRESPONSE
DESCRIPTOR.message_types_by_name['UnsubscribeRequest'] = _UNSUBSCRIBEREQUEST
DESCRIPTOR.message_types_by_name['CloseRequest'] = _CLOSEREQUEST
DESCRIPTOR.message_types_by_name['CloseResponse'] = _CLOSERESPONSE
DESCRIPTOR.enum_types_by_name['StartPosition'] = _STARTPOSITION

PubMsg = _reflection.GeneratedProtocolMessageType('PubMsg', (_message.Message,), dict(
  DESCRIPTOR = _PUBMSG,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.PubMsg)
  ))
_sym_db.RegisterMessage(PubMsg)

PubAck = _reflection.GeneratedProtocolMessageType('PubAck', (_message.Message,), dict(
  DESCRIPTOR = _PUBACK,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.PubAck)
  ))
_sym_db.RegisterMessage(PubAck)

MsgProto = _reflection.GeneratedProtocolMessageType('MsgProto', (_message.Message,), dict(
  DESCRIPTOR = _MSGPROTO,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.MsgProto)
  ))
_sym_db.RegisterMessage(MsgProto)

Ack = _reflection.GeneratedProtocolMessageType('Ack', (_message.Message,), dict(
  DESCRIPTOR = _ACK,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.Ack)
  ))
_sym_db.RegisterMessage(Ack)

ConnectRequest = _reflection.GeneratedProtocolMessageType('ConnectRequest', (_message.Message,), dict(
  DESCRIPTOR = _CONNECTREQUEST,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.ConnectRequest)
  ))
_sym_db.RegisterMessage(ConnectRequest)

ConnectResponse = _reflection.GeneratedProtocolMessageType('ConnectResponse', (_message.Message,), dict(
  DESCRIPTOR = _CONNECTRESPONSE,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.ConnectResponse)
  ))
_sym_db.RegisterMessage(ConnectResponse)

SubscriptionRequest = _reflection.GeneratedProtocolMessageType('SubscriptionRequest', (_message.Message,), dict(
  DESCRIPTOR = _SUBSCRIPTIONREQUEST,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.SubscriptionRequest)
  ))
_sym_db.RegisterMessage(SubscriptionRequest)

SubscriptionResponse = _reflection.GeneratedProtocolMessageType('SubscriptionResponse', (_message.Message,), dict(
  DESCRIPTOR = _SUBSCRIPTIONRESPONSE,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.SubscriptionResponse)
  ))
_sym_db.RegisterMessage(SubscriptionResponse)

UnsubscribeRequest = _reflection.GeneratedProtocolMessageType('UnsubscribeRequest', (_message.Message,), dict(
  DESCRIPTOR = _UNSUBSCRIBEREQUEST,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.UnsubscribeRequest)
  ))
_sym_db.RegisterMessage(UnsubscribeRequest)

CloseRequest = _reflection.GeneratedProtocolMessageType('CloseRequest', (_message.Message,), dict(
  DESCRIPTOR = _CLOSEREQUEST,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.CloseRequest)
  ))
_sym_db.RegisterMessage(CloseRequest)

CloseResponse = _reflection.GeneratedProtocolMessageType('CloseResponse', (_message.Message,), dict(
  DESCRIPTOR = _CLOSERESPONSE,
  __module__ = 'protocol_pb2'
  # @@protoc_insertion_point(class_scope:pb.CloseResponse)
  ))
_sym_db.RegisterMessage(CloseResponse)


# @@protoc_insertion_point(module_scope)



from setuptools import setup
import re
import ast


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('stan/aio/client.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='asyncio-nats-streaming',
    version=version,
    description='NATS Streaming client for Python Asyncio',
    long_description='Asyncio based Python client for NATS Streaming',
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
        ],
    url='https://github.com/nats-io/stan.py',
    author='Waldemar Quevedo',
    author_email='wally@synadia.com',
    license='Apache 2 License',
    packages=['stan', 'stan.aio', 'stan.pb'],
    zip_safe=True,
    install_requires=['protobuf>=3.4','asyncio-nats-client>=0.7.0'],
)
